//! Handle caching of files, specifically book covers.

use std::{
    fs::{self, File},
    path::{Path, PathBuf},
};

use sha2::{
    digest::{generic_array::GenericArray, typenum::U32},
    Digest, Sha256,
};
use std::fmt::Write;
use thiserror::Error;
use tracing::debug;

/// Errors from dealing with file caching.
#[derive(Error, Debug)]
pub enum CacheError {
    /// Error converting a hash to its string representation.
    #[error("failed to access thumbnail")]
    HashError(#[from] std::fmt::Error),
    /// Error creating a thumbnail for an image..
    #[error("failed to create thumbnail")]
    ImageError(#[from] image::ImageError),
    /// Error accessing a thumbnail.
    #[error("failed to access thumbnail")]
    ThumbnailAccessError(#[from] std::io::Error),
    /// Error accessing thumbnail directories.
    #[error("failed to access thumbnail directory")]
    ThumbnailPathError(PathBuf),
}

/// Convert a hash into its path representation inside the cache directory.
///
/// First hash character is the top folder, second character the second level folder and the rest
/// is the filename.
fn hash_to_path(hash: GenericArray<u8, U32>, cache_path: &Path) -> Result<PathBuf, CacheError> {
    let mut hash_string = String::new();
    for byte in hash {
        write!(&mut hash_string, "{:02x}", byte)?;
    }
    let hash = hash_string;

    let first_segment = &hash[0..1];
    let second_segment = &hash[1..2];
    let remaining_segment = &hash[2..];

    Ok(PathBuf::from(cache_path)
        .join(first_segment)
        .join(second_segment)
        .join(remaining_segment))
}

/// Create a thumbnail for `cover_path` at `thumbnail_path`.
fn create_thumbnail(cover_path: &Path, thumbnail_path: &Path) -> Result<(), CacheError> {
    debug!("creating thumbnail for {}", cover_path.to_string_lossy());
    let folders = thumbnail_path
        .parent()
        .ok_or_else(|| CacheError::ThumbnailPathError(thumbnail_path.to_path_buf()))?;
    fs::create_dir_all(folders)?;

    const THUMBNAIL_SIZE: u32 = 512;
    let img = image::open(cover_path)?;
    let thumbnail = img.thumbnail(THUMBNAIL_SIZE, THUMBNAIL_SIZE);
    thumbnail.save_with_format(thumbnail_path, image::ImageFormat::Jpeg)?;
    debug!("saved thumbnail to {}", thumbnail_path.to_string_lossy());

    Ok(())
}

/// Get the thumbnail for a book cover.
///
/// If a thumbnail does not yet exist, create it.
pub fn get_thumbnail(cover_path: &Path, cache_path: &Path) -> Result<File, CacheError> {
    let path_str = cover_path.to_string_lossy();
    let mut hasher = Sha256::new();
    hasher.update(path_str.as_bytes());
    let hash = hasher.finalize();

    let thumbnail_path = hash_to_path(hash, cache_path)?;

    if !thumbnail_path.exists() {
        create_thumbnail(cover_path, &thumbnail_path)?;
    }

    Ok(File::open(thumbnail_path)?)
}
