//! Enrich the [`Book`](struct@calibre_db::data::book::Book) type with additional information.

use std::{collections::HashMap, fmt::Display, path::Path};

use calibre_db::data::{
    author::Author as DbAuthor, book::Book as DbBook, series::Series as DbSeries,
};
use serde::Serialize;

use crate::app_state::AppState;

/// Wrapper type for a file format string (must be a struct in order to implement traits).
#[derive(Debug, Clone, Serialize, Eq, PartialEq, Hash)]
pub struct Format(pub String);
/// Wrapper type for a collection of formats, [`Format`](struct@Format) on the left and a String
/// signifying its file path on the right.
pub type Formats = HashMap<Format, String>;

/// Recognize `pdf` and `epub` and return their value, everything else transforms to `unknown`.
impl Display for Format {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self.0.as_ref() {
            "pdf" => write!(f, "pdf"),
            "epub" => write!(f, "epub"),
            _ => write!(f, "unknown"),
        }
    }
}

/// Wrapper around [`Book`](struct@calibre_db::data::book::Book) from the
/// [`calibre-db`](mod@calibre_db) crate. Adding
/// [`Author`](struct@calibre_db::data::author::Author),
/// [`Series`](struct@calibre_db::data::series::Series) as well as
/// [`Format`](struct@Format) information.
#[derive(Debug, Clone, Serialize)]
pub struct Book {
    /// Book data from the database.
    pub data: DbBook,
    /// Author information.
    pub author: DbAuthor,
    /// Series information. Series on the left and the index of the book within
    /// that series on the right.
    pub series: Option<(DbSeries, f64)>,
    /// Format information.
    pub formats: Formats,
}

impl Book {
    /// Wrap a [`DbBook`](struct@calibre_db::data::book::Book) in a [`Book`](struct@Book) and
    /// enrich it with additional information.
    ///
    /// `db_series` is an Option tuple with the series on the left and the index of the book within
    /// that series on the right.
    pub fn from_db_book(
        db_book: &DbBook,
        db_series: Option<(DbSeries, f64)>,
        author: DbAuthor,
        formats: Formats,
    ) -> Self {
        Self {
            data: db_book.clone(),
            author: author.clone(),
            series: db_series.map(|x| (x.0, x.1)),
            formats,
        }
    }

    /// Find all pdf and epub formats of a book on disk. Search their directory in the library for
    /// it.
    fn formats(book: &DbBook, library_path: &Path) -> Formats {
        let book_path = library_path.join(&book.path);
        let mut formats = HashMap::new();

        for entry in book_path.read_dir().unwrap().flatten() {
            if let Some(extension) = entry.path().extension() {
                let format = match extension.to_string_lossy().to_string().as_str() {
                    "pdf" => Some("pdf".to_string()),
                    "epub" => Some("epub".to_string()),
                    _ => None,
                };
                if let Some(format) = format {
                    formats.insert(
                        Format(format),
                        entry.file_name().to_string_lossy().to_string(),
                    );
                }
            }
        }

        formats
    }

    /// Wrap a [`DbBook`](struct@calibre_db::data::book::Book) in a [`Book`](struct@Book) by
    /// fetching additional information about author, formats and series.
    pub fn full_book(book: &DbBook, state: &AppState) -> Option<Book> {
        let formats = Book::formats(book, &state.config.library_path);
        let author = state.calibre.book_author(book.id).ok()?;
        let series = state.calibre.book_series(book.id).ok()?;
        Some(Book::from_db_book(book, series, author, formats))
    }
}
