//! Handle requests for multiple books.

use std::sync::Arc;

use calibre_db::data::pagination::SortOrder;
use poem::{
    error::NotFoundError,
    handler,
    web::{Data, Path},
    Response,
};
use tokio::fs::File;

use crate::{
    app_state::AppState,
    data::book::{Book, Format},
    handlers::error::HandlerError,
    opds::media_type::MediaType,
    Accept,
};

/// Handle a request for multiple books, starting at the first.
#[handler]
pub async fn handler_init(
    accept: Data<&Accept>,
    state: Data<&Arc<AppState>>,
) -> Result<Response, poem::Error> {
    books(&accept, &state, None, &SortOrder::ASC).await
}

/// Handle a request for multiple books, starting at the `cursor` and going in the direction of
/// `sort_order`.
#[handler]
pub async fn handler(
    Path((cursor, sort_order)): Path<(String, SortOrder)>,
    accept: Data<&Accept>,
    state: Data<&Arc<AppState>>,
) -> Result<Response, poem::Error> {
    books(&accept, &state, Some(&cursor), &sort_order).await
}

/// Handle a request for a book with id `id` in format `format`.
#[handler]
pub async fn handler_download(
    Path((id, format)): Path<(u64, String)>,
    state: Data<&Arc<AppState>>,
) -> Result<Response, poem::Error> {
    let book = state
        .calibre
        .scalar_book(id)
        .map_err(HandlerError::DataError)?;
    let book = Book::full_book(&book, &state).ok_or(NotFoundError)?;
    let format = Format(format);
    let file_name = book.formats.get(&format).ok_or(NotFoundError)?;
    let file_path = state
        .config
        .library_path
        .join(book.data.path)
        .join(file_name);
    let mut file = File::open(file_path).await.map_err(|_| NotFoundError)?;
    let content_type: MediaType = format.into();
    let content_type = format!("{content_type}");

    crate::handlers::download::handler(file_name, file, &content_type).await
}

async fn books(
    accept: &Accept,
    state: &Arc<AppState>,
    cursor: Option<&str>,
    sort_order: &SortOrder,
) -> Result<Response, poem::Error> {
    match accept {
        Accept::Html => crate::handlers::html::books::handler(state, cursor, sort_order).await,
        Accept::Opds => crate::handlers::opds::books::handler(state, cursor, sort_order).await,
    }
}
