//! Handle requests for cover images.

use std::{fs::File, path::Path as FilePath, sync::Arc};

use crate::{
    app_state::AppState,
    cache::{self, CacheError},
    config::Config,
    handlers::error::HandlerError,
};
use calibre_db::calibre::Calibre;
use poem::{
    error::NotFoundError,
    handler,
    web::{headers::ContentType, Data, Path},
    Response,
};
use thiserror::Error;
use tokio::fs::File as AsyncFile;

/// Errors from fetching cover images.
#[derive(Error, Debug)]
pub enum CoverError {
    /// Error fetching a cover thumbnail.
    #[error("failed to access thumbnail")]
    ThumbnailError(#[from] CacheError),
    /// Error fetching a full cover.
    #[error("failed access cover")]
    FullCoverError(#[from] std::io::Error),
}

/// Handle a request for the cover thumbnail of book with id `id`.
#[handler]
pub async fn handler_thumbnail(
    id: Path<u64>,
    state: Data<&Arc<AppState>>,
) -> Result<Response, poem::Error> {
    cover(
        &state.calibre,
        &state.config,
        *id,
        |cover_path, cache_path| Ok(cache::get_thumbnail(cover_path, cache_path)?),
    )
    .await
}

/// Handle a request for the cover image of book with id `id`.
#[handler]
pub async fn handler_full(
    id: Path<u64>,
    state: Data<&Arc<AppState>>,
) -> Result<Response, poem::Error> {
    cover(&state.calibre, &state.config, *id, |cover_path, _| {
        Ok(File::open(cover_path)?)
    })
    .await
}

async fn cover<F>(
    calibre: &Calibre,
    config: &Config,
    id: u64,
    f: F,
) -> Result<Response, poem::Error>
where
    F: Fn(&FilePath, &FilePath) -> Result<File, CoverError>,
{
    let book = calibre.scalar_book(id).map_err(HandlerError::DataError)?;
    let cover_path = config.library_path.join(book.path).join("cover.jpg");

    let cover = f(&cover_path, &config.cache_path).map_err(|_| NotFoundError)?;
    let cover = AsyncFile::from_std(cover);
    crate::handlers::download::handler("cover.jpg", cover, &ContentType::jpeg().to_string()).await
}
