//! Handle multiple authors for opds.

use calibre_db::{
    calibre::Calibre,
    data::{author::Author as DbAuthor, pagination::SortOrder},
};
use poem::{IntoResponse, Response};
use time::OffsetDateTime;

use crate::{
    handlers::error::HandlerError,
    opds::{entry::Entry, feed::Feed, link::Link, media_type::MediaType, relation::Relation},
    APP_NAME,
};

/// Render all authors as OPDS entries embedded in a feed.
pub async fn handler(
    calibre: &Calibre,
    _cursor: Option<&str>,
    _sort_order: &SortOrder,
) -> Result<Response, poem::Error> {
    let authors: Vec<DbAuthor> = calibre
        .authors(u32::MAX.into(), None, &SortOrder::ASC)
        .map_err(HandlerError::DataError)?;

    let entries: Vec<Entry> = authors.into_iter().map(Entry::from).collect();
    let now = OffsetDateTime::now_utc();

    let self_link = Link {
        href: "/opds/authors".to_string(),
        media_type: MediaType::Navigation,
        rel: Relation::Myself,
        title: None,
        count: None,
    };
    let feed = Feed::create(
        now,
        &format!("{APP_NAME}:authors"),
        "All Authors",
        self_link,
        vec![],
        entries,
    );
    let xml = feed.as_xml().map_err(HandlerError::OpdsError)?;

    Ok(xml
        .with_content_type("application/atom+xml")
        .into_response())
}
