//! Handle multiple books for opds.

use calibre_db::data::pagination::SortOrder;
use poem::{IntoResponse, Response};
use time::OffsetDateTime;

use crate::{
    app_state::AppState,
    data::book::Book,
    handlers::error::HandlerError,
    opds::{entry::Entry, feed::Feed, link::Link, media_type::MediaType, relation::Relation},
    APP_NAME,
};

/// Render all books as OPDS entries embedded in a feed.
pub async fn handler(
    state: &AppState,
    _cursor: Option<&str>,
    _sort_order: &SortOrder,
) -> Result<Response, poem::Error> {
    let books: Vec<Book> = state
        .calibre
        .books(u32::MAX.into(), None, &SortOrder::ASC)
        .map(|x| x.iter().filter_map(|y| Book::full_book(y, state)).collect())
        .map_err(HandlerError::DataError)?;
    render_books(books).await
}

/// Render a list of books as OPDS entries in a feed.
pub(crate) async fn render_books(books: Vec<Book>) -> Result<Response, poem::Error> {
    let entries: Vec<Entry> = books.into_iter().map(Entry::from).collect();
    let now = OffsetDateTime::now_utc();

    let self_link = Link {
        href: "/opds/books".to_string(),
        media_type: MediaType::Navigation,
        rel: Relation::Myself,
        title: None,
        count: None,
    };
    let feed = Feed::create(
        now,
        &format!("{APP_NAME}:books"),
        "All Books",
        self_link,
        vec![],
        entries,
    );
    let xml = feed.as_xml().map_err(HandlerError::OpdsError)?;

    Ok(xml
        .with_content_type("application/atom+xml")
        .into_response())
}
