//! A very simple ebook server for a calibre library, providing a html interface as well as an OPDS feed.
//!
//! Shamelessly written to scratch my own itches.

use std::sync::Arc;

use app_state::AppState;
use calibre_db::calibre::Calibre;
use config::Config;
use poem::{
    endpoint::EmbeddedFilesEndpoint, get, listener::TcpListener, middleware::Tracing, EndpointExt,
    Route, Server,
};
use rust_embed::RustEmbed;
use tokio::signal;
use tracing::info;

pub mod app_state;
pub mod cache;
pub mod cli;
pub mod config;
/// Data structs and their functions.
pub mod data {
    pub mod book;
}
/// Request handlers. Because it can not be guaranteed that a proper accept header is sent, the
/// routes are doubled and the decision on whether to render html or OPDS is made with internal
/// data on the respective routes.
pub mod handlers {
    /// Handle requests for html.
    pub mod html {
        pub mod author;
        pub mod authors;
        pub mod books;
        pub mod recent;
        pub mod search;
        pub mod series;
        pub mod series_single;
    }
    /// Handle requests for OPDS.
    pub mod opds {
        pub mod author;
        pub mod authors;
        pub mod books;
        pub mod feed;
        pub mod recent;
        pub mod search;
        pub mod search_info;
        pub mod series;
        pub mod series_single;
    }
    pub mod author;
    pub mod authors;
    pub mod books;
    pub mod cover;
    pub mod download;
    pub mod error;
    pub mod paginated;
    pub mod recent;
    pub mod search;
    pub mod series;
    pub mod series_single;
    pub mod source_archive;
}
/// OPDS data structs.
pub mod opds {
    pub mod author;
    pub mod content;
    pub mod entry;
    pub mod error;
    pub mod feed;
    pub mod link;
    pub mod media_type;
    pub mod relation;
    pub mod search;
}
pub mod templates;

pub const APP_NAME: &str = "little-hesinde";
pub const VERSION: &str = "0.3.1";

/// Internal marker data in lieu of a proper `Accept` header.
#[derive(Debug, Clone, Copy)]
pub enum Accept {
    /// Render as html.
    Html,
    /// Render as OPDS.
    Opds,
}

/// Embedd static files.
#[derive(RustEmbed)]
#[folder = "static"]
pub struct Files;

/// Main entry point to run the ebook server with a calibre library specified in `config`.
pub async fn run(config: Config) -> Result<(), std::io::Error> {
    let calibre = Calibre::load(&config.metadata_path).expect("failed to load calibre database");
    let app_state = Arc::new(AppState {
        calibre,
        config: config.clone(),
    });

    let html_routes = Route::new()
        .at("/", get(handlers::recent::handler))
        .at("/books", get(handlers::books::handler_init))
        .at("/books/:cursor/:sort_order", get(handlers::books::handler))
        .at("/series", get(handlers::series::handler_init))
        .at(
            "/series/:cursor/:sort_order",
            get(handlers::series::handler),
        )
        .at("/series/:id", get(handlers::series_single::handler))
        .at("/authors", get(handlers::authors::handler_init))
        .at("/authors/:id", get(handlers::author::handler))
        .at(
            "/authors/:cursor/:sort_order",
            get(handlers::authors::handler),
        )
        .at("/cover/:id", get(handlers::cover::handler_full))
        .at(
            "/cover/:id/thumbnail",
            get(handlers::cover::handler_thumbnail),
        )
        .at("/book/:id/:format", get(handlers::books::handler_download))
        .at("/archive", get(handlers::source_archive::handler))
        .at("/search", get(handlers::search::handler))
        .nest("/static", EmbeddedFilesEndpoint::<Files>::new())
        .data(Accept::Html);

    let opds_routes = Route::new()
        .at("/", get(handlers::opds::feed::handler))
        .at("/recent", get(handlers::recent::handler))
        .at("/books", get(handlers::books::handler_init))
        .at("/authors", get(handlers::authors::handler_init))
        .at("/authors/:id", get(handlers::author::handler))
        .at("/series", get(handlers::series::handler_init))
        .at("/series/:id", get(handlers::series_single::handler))
        .at("/search/info", get(handlers::opds::search_info::handler))
        .at("/search", get(handlers::search::handler))
        .data(Accept::Opds);

    let app = Route::new()
        .nest("/", html_routes)
        .nest("/opds", opds_routes)
        .data(app_state)
        .with(Tracing);

    let server = Server::new(TcpListener::bind(config.listen_address))
        .name("little-hesinde")
        .run(app);

    tokio::select! {
        _ = server => {},
        _ = signal::ctrl_c() => {
            info!("Received Ctrl+C, shutting down...");
        },
    }
    Ok(())
}
