//! Root feed data.

use std::io::Cursor;

use quick_xml::{
    events::{BytesDecl, BytesStart, Event},
    se::to_string,
    Reader, Writer,
};
use serde::Serialize;
use time::OffsetDateTime;

use super::{
    author::Author, entry::Entry, error::OpdsError, link::Link, media_type::MediaType,
    relation::Relation,
};

/// Root feed element of OPDS.
#[derive(Debug, Serialize)]
#[serde(rename = "feed")]
pub struct Feed {
    /// Title, often shown in OPDS clients.
    pub title: String,
    /// Feed id.
    pub id: String,
    /// When was the feed updated last.
    #[serde(with = "time::serde::rfc3339")]
    pub updated: OffsetDateTime,
    /// Icon for the feed.
    pub icon: String,
    /// Feed author.
    pub author: Author,
    /// Links, for example home or self.
    #[serde(rename = "link")]
    pub links: Vec<Link>,
    /// Entries inside the feed (books, series, subsections, ...)
    #[serde(rename = "entry")]
    pub entries: Vec<Entry>,
}

impl Feed {
    /// Create a feed with the specified data.
    pub fn create(
        now: OffsetDateTime,
        id: &str,
        title: &str,
        self_link: Link,
        mut additional_links: Vec<Link>,
        entries: Vec<Entry>,
    ) -> Self {
        let author = Author {
            name: "Thallian".to_string(),
            uri: "https://code.vanwa.ch/shu/little-hesinde".to_string(),
            email: None,
        };
        let mut links = vec![
            Link {
                href: "/opds".to_string(),
                media_type: MediaType::Navigation,
                rel: Relation::Start,
                title: Some("Home".to_string()),
                count: None,
            },
            Link {
                href: "/opds/search/info".to_string(),
                media_type: MediaType::Search,
                rel: Relation::Search,
                title: Some("Search".to_string()),
                count: None,
            },
            self_link,
        ];
        links.append(&mut additional_links);

        Self {
            title: title.to_string(),
            id: id.to_string(),
            updated: now,
            icon: "favicon.ico".to_string(),
            author,
            links,
            entries,
        }
    }

    /// Serialize a feed to OPDS xml.
    pub fn as_xml(&self) -> Result<String, OpdsError> {
        let xml = to_string(&self)?;
        let mut reader = Reader::from_str(&xml);
        reader.config_mut().trim_text(true);

        let declaration = BytesDecl::new("1.0", Some("UTF-8"), None);
        let mut writer = Writer::new(Cursor::new(Vec::new()));
        writer.write_event(Event::Decl(declaration))?;

        let mut feed_start = BytesStart::new("feed");
        feed_start.push_attribute(("xmlns", "http://www.w3.org/2005/Atom"));
        feed_start.push_attribute(("xmlns:xhtml", "http://www.w3.org/1999/xhtml"));
        feed_start.push_attribute(("xmlns:opds", "http://opds-spec.org/2010/catalog"));
        feed_start.push_attribute(("xmlns:opensearch", "http://a9.com/-/spec/opensearch/1.1/"));
        feed_start.push_attribute(("xmlns:dcterms", "http://purl.org/dc/terms/"));
        feed_start.push_attribute(("xmlns:thr", "http://purl.org/syndication/thread/1.0"));

        loop {
            match reader.read_event() {
                Ok(Event::Start(e)) if e.name().as_ref() == b"feed" => {
                    writer.write_event(Event::Start(feed_start.clone()))?
                }
                Ok(Event::Eof) => break,
                Ok(e) => writer.write_event(e)?,
                Err(e) => return Err(e)?,
            }
        }
        let result = writer.into_inner().into_inner();
        Ok(String::from_utf8(result)?)
    }
}
