//! Media types for OPDS elements.

use serde_with::SerializeDisplay;

use crate::data::book::Format;

/// Valid media types.
#[derive(Debug, Copy, Clone, SerializeDisplay)]
pub enum MediaType {
    /// A link with this type is meant to acquire a certain thing, for example an entry.
    Acquisition,
    Epub,
    Html,
    Jpeg,
    /// A link with this type is meant for navigation around a feed.
    Navigation,
    Pdf,
    Text,
    Search,
}

/// Convert `epub` and `pdf` formats to their respective media type. Everything else is `Text`.
impl From<Format> for MediaType {
    fn from(value: Format) -> Self {
        match value.0.as_ref() {
            "epub" => MediaType::Epub,
            "pdf" => MediaType::Pdf,
            _ => MediaType::Text,
        }
    }
}

/// Display the respective mime types of the respective media types.
impl std::fmt::Display for MediaType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            MediaType::Acquisition => write!(
                f,
                "application/atom+xml;profile=opds-catalog;kind=aqcuisition"
            ),
            MediaType::Epub => write!(f, "application/epub+zip"),
            MediaType::Html => write!(f, "text/html"),
            MediaType::Jpeg => write!(f, "image/jpeg"),
            MediaType::Navigation => write!(
                f,
                "application/atom+xml;profile=opds-catalog;kind=navigation"
            ),
            MediaType::Pdf => write!(f, "application/pdf"),
            MediaType::Text => write!(f, "text"),
            MediaType::Search => write!(f, "application/opensearchdescription+xml"),
        }
    }
}
